import sys
import json
import re

import ssl
import urllib
from http.client import HTTPSConnection, HTTPResponse
from requests import Session
from requests.auth import HTTPBasicAuth

from io import StringIO
from lxml import etree as ET

import pymex.pypsic

class Client():

    def __init__( self, mode='dev', debug=False ):
        self.mode = mode
        self.debug = debug

        self.ns = { 'reg' :"http://hupo.psi.org/psicquic/registry" }
        self.registryUrl = "http://www.ebi.ac.uk/Tools/webservices/psicquic/registry/registry?action=STATUS&format=xml"
        self.services = None
        self.default = "IMEx"

    @property                   
    def registry( self ):

        if self.services is not None:
            return self.services
                
        try:
            scontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            with urllib.request.urlopen( self.registryUrl,
                                         context=scontext ) as f:

                record = f.read().decode('utf-8')
                record = re.sub(r'<\?xml[^>]+>', '', record)

                parser = ET.XMLParser( remove_blank_text=True )
                dom = ET.parse(StringIO(record), parser)

                slst = dom.xpath( '//reg:service', namespaces = self.ns )

                services = {}
                
                for s in slst:
                    srv = {}
                    srv['name'] = s.xpath( './reg:name/text()',
                                           namespaces = self.ns )[0]
                    srv['rurl'] = s.xpath( './reg:restUrl/text()',
                                           namespaces = self.ns )[0]
                    srv['active'] = s.xpath( './reg:active/text()',
                                             namespaces = self.ns )[0]                    
                                        
                    services[ srv['name'].lower() ] = srv                    
                self.services = services
                
        except Exception as e:
            print( e )
            return None
        
        return pypsic.Registry( self.services )
        
    def getinteractor( self, query, server="DEFAULT",
                       first = 0, max = 10, format="xml25" ):
         
        if self.services is None:
            print("#Getting registry")
            self.registry

        if self.services is None:
            sys.exit(1)
            
        if server in ['DEFAULT']:        
            server = self.default

        print("#PSICQUC Server: " + server )
        
        if self.services is None or server.lower() not in self.services:
            sys.exit(1)

        if self.services[server.lower()]['active'].lower() == 'false':
            print("#PSICQUC Server not active")
            sys.exit(1)

        baseUrl = self.services[server.lower()]['rurl']    
        try:
            url = baseUrl +'interactor/' + query
            par = {'firstResult': first, 'maxResults': max, 'format': format}
            urlpar = urllib.parse.urlencode( par )            
            
            scontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            with urllib.request.urlopen( url + "?" + urlpar,
                                         context=scontext) as f:
                record = f.read().decode('utf-8')

            return record

        except Exception as e:
             print( e )
                   
        return None

    def getinteraction( self, query, server="DEFAULT",
                        first = 0, max = 10, format="xml25" ):
        
        if self.services is None:
            print("#Getting registry")
            self.registry

        if server in ['DEFAULT']:        
            server = self.default

        print("###PSICQUC Server: " + server )
        
        if self.services is None or server.lower() not in self.services:
            sys.exit(1)

        if self.services[server.lower()]['active'].lower() == 'false':
            print("#PSICQUC Server not active")
            sys.exit(1)

        baseUrl = self.services[server.lower()]['rurl']    
        
        try:

            url = baseUrl +'interaction/' + query
            par = {'firstResult': first, 'maxResults': max, 'format': format}
            urlpar = urllib.parse.urlencode( par )            

            print(url + "?" + urlpar)
            scontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            with urllib.request.urlopen( url + "?" + urlpar,
                                         context=scontext) as f:
                record = f.read().decode('utf-8')

            return record

        except Exception as e:
             print( e )
                   
        return None

    def getquery( self, query, server="DEFAULT",
                  first = 0, max = 10, format="xml25" ):

        if self.services is None:
            print("##Getting registry")
            self.registry
            
        if server in ['DEFAULT']:        
            server = self.default

        print("#PSICQUC Server: " + server )
         
        if self.services is None or server.lower() not in self.services:
            sys.exit(1)

        if self.services[server.lower()]['active'].lower() == 'false':
            print("#PSICQUC Server not active")
            sys.exit(1)

        baseUrl = self.services[server.lower()]['rurl']    
        
        try:

            url = baseUrl +'query/' + urllib.parse.quote(query)             
            par = {'firstResult': first, 'maxResults': max, 'format': format}
            urlpar = urllib.parse.urlencode( par )            
            print(url + "?" + urlpar)
            scontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            with urllib.request.urlopen( url + "?" + urlpar,
                                         context=scontext ) as f:
                record = f.read().decode('utf-8')

            return record

        except Exception as e:
             print( e )
                   
        return None



import json
import sys
import os
import re
from lxml import etree
from urllib.request import urlopen

import pymex

class RecordBuilder():

    def __init__(self,debug=False):
        self.debug = debug
        self.oboUrl = 'https://www.ebi.ac.uk/ols/api/ontologies/%CVT%/terms?iri=http://purl.obolibrary.org/obo/'
        self.taxUrl = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=taxonomy&retmode=xml&id='
        self.uniNs = { "up":"http://uniprot.org/uniprot" }
        self.uniUrl = 'https://www.uniprot.org/uniprot/'
        self.cvreg = re.compile("([^:]+:\d+)(\(.+\))?")
        self.cvtrec = {}
        self.taxrec = {}
        self.unirec = {}
        self.ebirec = {}

        self.feature = False 
        
        self.template = {}
        self.cvdict = {}

        self.root = []
        self.pdir = os.path.dirname( os.path.realpath(__file__) )
        
        srcpath = os.path.join( self.pdir,'jmif-source.jmif' )
        self.template["source"] = json.load( open( srcpath, 'r' ) )["source"]

        cvtpath = os.path.join( self.pdir,'jmif-cvterm.jmif' )
        self.cvdict = json.load( open( cvtpath, 'r' ) )

    def cvterm( self, cvid ):
        
        cvmatch = self.cvreg.match( cvid )
        if cvmatch:
            cvid = cvmatch.group(1)
            cvlabel = None
        else:
            try:
                cvid = self.cvdict[cvid]
                cvlabel = None
            except:
                cvlabel = cvid
                cvid = None
            
        if cvlabel is not None:
            term = {"label" : cvlabel }
            return term
        
        if cvid not in self.cvtrec:

            #fetch term from OLS

            cvc = cvid.split(":")
            
            cvurl = self.oboUrl.replace("%CVT%",cvc[0].lower()) + cvid.replace(':','_')            
            jcv = json.load( urlopen( cvurl ) )
            
            term = {}
            term['label'] = jcv['_embedded']['terms'][0]['label']
            
            if "description" in jcv['_embedded']['terms'][0]:
                 term['def'] = jcv['_embedded']['terms'][0]["description"]
                 
            elif "definition" in jcv['_embedded']['terms'][0]['annotation']:
                term['def'] = jcv['_embedded']['terms'][0]['annotation']['definition'][0]
                
            term['id'] = jcv['_embedded']['terms'][0]['obo_id']

            try:
                term["dbac"]=self.cvdict[cvc[0].lower()]["dbac"]
                term["db"]=self.cvdict[cvc[0].lower()]["db"]
            except:
                term["dbac"]=None
                term["db"]=None
            self.cvtrec[ term['id'] ] = term 
        
        return self.cvtrec[cvid]

    def buildCvTerm( self, cvid ):
        
        term = self.cvterm( cvid )

        res = { "names": { "shortLabel": term["label"] } }

        if "id" in term:
            
            res["xref"] = { "primaryRef": self.buildXref( term["id"],
                                                          db=term["db"],
                                                          dbAc= term["dbac"]) }
        return res

    def buildXref( self, acc, db="uniprotkb", dbAc="MI:0486", version = None,
                   refType="identity", refTypeAc="MI:0356" ):
    
        res = { "id": acc, "refType": refType, "refTypeAc": refTypeAc }                    

        if db is not None:
            res["db"] = db
            
        if dbAc is not None:
            res["dbAc"] = dbAc 
        
        if version is not None: 
            res["version"] = version

        return res
    
    def taxon(self, taxid):  #  12234    1234:1234

        (exptx,spctx) = (taxid.split(":") + [""])[0:2]
        taxid = exptx
        
        if taxid in ["-1","in vitro"]:
            tax = {"lname":"in vitro","sname":"in vitro","taxid":"-1"}
            self.taxrec[ "-1" ] = tax
            taxid = "-1"
        if taxid not in self.taxrec:
            tax = {}

            taxurl = self.taxUrl + taxid
            
            taxpath = "/TaxaSet/Taxon[./TaxId/text()="+taxid+"]"
            record = etree.parse( urlopen( taxurl ) )
            trec = record.xpath(taxpath )

            if trec:
                tax['taxid'] = taxid 
                sname = trec[0].xpath("./OtherNames/CommonName/text()")
                lname = trec[0].xpath("./ScientificName/text()") 

                if lname:
                    tax['lname'] = lname[0]
            
                if sname:
                    tax['sname'] = sname[0]
                else:
                    tax['sname'] = tax['lname']
                
                self.taxrec[ taxid ] = tax
                
        return self.taxrec[ taxid ]
        
    def uniprot(self, acc):
        if acc not in self.unirec:
            uprot = {}
        
            uniurl = self.uniUrl + acc +".xml"
            
            accpath = "/up:uniprot/up:entry/up:accession/text()"
            verpath = "/up:uniprot/up:entry/@version"
            snamepath = "/up:uniprot/up:entry/up:name/text()"
            lnamepath = "/up:uniprot/up:entry/up:protein/up:recommendedName/up:fullName/text()"
            taxpath = "/up:uniprot/up:entry/up:organism/up:dbReference[./@type='NCBI Taxonomy']/@id"
            record = etree.parse( urlopen( uniurl ) )
        
            uacc = record.xpath( accpath, namespaces = self.uniNs )
            if uacc:
                uprot["acc"] = str( uacc[0] )

            version = record.xpath( verpath, namespaces = self.uniNs )
            if version:
                uprot["version"] = str( version[0] )
           
            sname = record.xpath( snamepath,namespaces = self.uniNs )            
            if sname:
                uprot["sname"] = str( sname[0] ) 
                                
            lname = record.xpath( lnamepath, namespaces = self.uniNs )
            if lname:
                uprot["lname"] = str( lname[0] ) 

            taxid = record.xpath( taxpath, namespaces = self.uniNs )
            if taxid:
                uprot["taxid"] = str( taxid[0]  )
                
            self.unirec[acc] = uprot
        
        return self.unirec[acc]


    def ebi(self, acc, taxid ):
        if acc not in self.unirec:
            ebi = {}

            ebi["acc"] = str( acc )
            ebi["version"] = str( "" )
            ebi["sname"] = str( "" ) 
            
            ebi["lname"] = str( "" ) 
            ebi["taxid"] = str( taxid )
                
            self.ebirec[acc] = ebi

            # db="intact" dbAc="MI:0469" 

            
        return self.ebirec[acc]
        
    def build( self, filename ):    
        
        record = {}
        source = None
        interaction = None
        interactor = None
        participant = None
        feature = None
        frange = None
        xtgt = None
        seqtgt = None
        bibref = None

        with open( filename, 'r' ) as sf:
            for ln in sf:
                ln = ln.strip()
                col = [s.strip() for s in ln.strip().split("\t")]
                
                if ln.startswith("source"):
                    self.feature = False
                    
                    record["source"] = {}
                    src = col[1]
                    if src in self.template["source"]:
                        record["source"] =  self.template["source"][src]
                    xtgt = record["source"]

                elif ln.startswith("pmid"):
                    pmid = col[1]
                    pref = self.buildXref( pmid, db="pubmed", dbAc="MI:0446",
                                           refType="primary-reference", refTypeAc="MI:0358" )    
                    bibref = {"xref":{"primaryRef": pref}}
                    
                elif ln.startswith("interaction"):
                    self.feature = False
                    interaction =  {"names":{"shortLabel": "N/A"},
                                    "experiment":[{"names":{"shortLabel": "N/A"}}]}
                    if bibref is not None:
                        interaction["experiment"][0]["bibref"] = bibref
                    
                    record.setdefault("interaction",[]).append(interaction)
                    
                    # interaction type 
                    interaction["interactionType"] = self.buildCvTerm(col[1])
                    
                    # interaction detection                 
                    interaction["experiment"][0]["interactionDetectionMethod"] = self.buildCvTerm(col[2])

                    # participant identification

                    interaction["experiment"][0]["participantIdentificationMethod"] = self.buildCvTerm("experimental particp") 
                    
                    # interaction host 

                    taxid =  col[3]                
                    ctax = self.taxon( taxid )

                    interaction["experiment"][0].setdefault("hostOrganism",[]).append( {                    
                        "names": {
                            "shortLabel": ctax["sname"],
                            "fullName": ctax["lname"] },
                        "ncbiTaxId": ctax["taxid"] } )
                    
                    if len(col) > 4:
                        ctype = self.buildCvTerm( col[4] )
                        if ctype is not None:
                            interaction["experiment"][0]["hostOrganism"][0]["cellType"] = ctype

                    if len(col) > 5:
                        compartnent = self.buildCvTerm( col[5] )
                        if compartnent is not None:
                            interaction["experiment"][0]["hostOrganism"][0]["compartment"] = compartnent

                    if len(col) > 6:
                        tissue = self.buildCvTerm( col[6] )
                        if tissue is not None:
                            interaction["experiment"][0]["hostOrganism"][0]["tissue"] = tissue
                        
                    xtgt = interaction    
                elif ln.startswith("molecule"):
                    self.feature = False
                    participant = {}
                    interaction.setdefault( "participant",[] ).append( participant )

                    # interactor 
                    interactor = {"xref":{} }
                    
                    participant["interactor"] = interactor 
                                                                                
                    if col[2].startswith("uprot:"):
                        (acc,ver) = (col[2].split(".") + [""])[0:2]                        
                        irec = self.uniprot( acc.replace("uprot:","") )
                        if len(ver) > 0:
                            irec["version"] = ver

                        interactor["xref"]["primaryRef"] = self.buildXref( irec["acc"],
                                                                           db="uniprotkb",
                                                                           dbAc= "MI:0486" )                            
                    elif col[2].startswith("ebi:"):
                        (acc,ver) = (col[2].split(".") + [""])[0:2]
                        (exptx,spctx) = (col[4].split(":") + [""])[0:2]
                                                
                        irec = self.ebi( acc.replace("ebi:",""), spctx )
                        if len(ver) > 0:
                            irec["version"] = ver

                        interactor["xref"]["primaryRef"] = self.buildXref( irec["acc"],
                                                                           db="intact",
                                                                           dbAc= "MI:0469" )                        
                    else:
                        pass                   
                    
                    # interactor names
                    if irec:
                        interactor["names"] = { "shortLabel":irec["sname"].lower(),
                                                "fullName":irec["lname"] }                                                            
                    
                    #interactor type
                    interactor["interactorType"] = self.buildCvTerm( col[1] )

                    # interactor organism                    
                    ctax = self.taxon( irec["taxid"] )
                                                           
                    interactor["organism"] = {                    
                        "names": {
                            "shortLabel": ctax["sname"],
                            "fullName": ctax["lname"] },
                        "ncbiTaxId": irec["taxid"] }

                    xtgt = participant["interactor"]
                    # participant names
                    participant["names"] = {"shortLabel":irec["sname"].lower()}
       
                    # host                     
                    if len(col) > 4:
                        hostOrganism = []
                        participant["hostOrganism"] = hostOrganism 

                        taxid =  col[4]
                        ctax = self.taxon( taxid )

                        hostOrganism.append( {
                            "names": {
                                "shortLabel": ctax["sname"],
                                "fullName": ctax["lname"] },
                            "ncbiTaxId": ctax["taxid"] } )

                        if len(col) > 5:
                            ctype = self.buildCvTerm( col[5] )
                            if ctype is not None:
                                hostOrganism[0]["cellType"] = ctype

                        if len(col) > 6:
                            compartnent = self.buildCvTerm( col[6] )
                            if compartnent is not None:
                                hostOrganism[0]["compartnent"] = compartnent

                        if len(col) > 7:
                            tissue = self.buildCvTerm( col[7] )
                            if tissue is not None:
                                hostOrganism[0]["tissue"] = tissue

                    #stoichiometry
                    catt = { "value":"Stoichiometry: " + col[3],
                             "name":"comment",
                             "nameAc":"MI:0612" }
                    participant.setdefault( "attribute", [] ).append(catt)                                                

                    # seq target

                    seqtgt = participant["interactor"]
                    
                elif ln.startswith("exprole"):
                    self.feature = False            
                    participant.setdefault("experimentalRole",[])
                    
                    for role in col[1:]:                                            
                        participant["experimentalRole"].append( self.buildCvTerm( role ) )

                elif ln.startswith("expprep"):                    
                    self.feature = False            
                    participant.setdefault("experimentalPreparation",[])                    
                    for prep in col[1:]:                                                
                        participant["experimentalPreparation"].append( self.buildCvTerm( prep ) )                    
                elif ln.startswith("biorole"):
                    self.feature = False            
                    participant.setdefault("biologicalRole",[])
                    
                    for role in col[1:]:                    
                        participant["biologicalRole"].append( self.buildCvTerm( role ) )
                        
                elif ln.startswith("idmethod"):                                        
                    if self.feature:                        
                        mthCV = self.buildCvTerm( col[1].strip() )
                        feature["featureDetectionMethod"] = mthCV
                    else:
                        participant.setdefault("participantIdentificationMethod",[])
                        for mth in col[1:]:
                            mthCV = self.buildCvTerm( mth )
                            participant["participantIdentificationMethod"].append( mthCV )
                                                
                elif ln.startswith("feature"):
                    self.feature = True
                    feature = {}
                    participant.setdefault("feature",[]).append( feature )

                    feature["featureType"] = self.buildCvTerm( col[1] )

                    if len(col) > 2:
                        feature["names"] = {}
                        feature["names"]["shortLabel"] = col[2]
                        if len(col) > 3:
                            feature["names"]["fullName"] = col[3]
                        else:
                            pass
                    else:
                        pass
                    
                    xtgt = feature
                    
                elif ln.startswith("range"):
                    self.feature = True
                    if "range" not in feature:
                        feature["featureRange"] = []
                    frange = {}
                    feature["featureRange"].append(frange)

                    seqtgt = frange

                    rbegin = col[1]
                    rend = col[2]

                    if '..' in rbegin:
                        frange["startStatus"] = self.buildCvTerm( "MI:0338" ) #range
                        (sbeg,ebeg) = rbegin.split('..')
                        frange["begin"]={"start": sbeg , "end": ebeg}
                        
                    elif rbegin in ['n','c']:
                        if rbegin == 'n':
                            frange["startStatus"] = self.buildCvTerm( "MI:1040" ) # n-term range                           
                        else:
                            frange["startStatus"] = self.buildCvTerm( "MI:1039" ) # c-term range
                    else:
                        frange["startStatus"] = self.buildCvTerm( "MI:0335" ) # certain
                        frange["begin"]={ "position": rbegin }
                    
                    if '..' in rend:
                        frange["endStatus"] = self.buildCvTerm( "MI:0338" ) #range
                        (sbeg,ebeg) = rbegin.split('..')
                        frange["end"]={"start": sbeg , "end": ebeg}

                    elif rend in ['n','c']:
                        #frange["end"]={ "position": "0" }

                        if rend == 'n':
                            frange["endStatus"] = self.buildCvTerm( "MI:1040" ) # n-term range                           
                        else:
                            frange["endStatus"] = self.buildCvTerm( "MI:1039" ) # c-term range
                    else:
                        frange["endStatus"] = self.buildCvTerm( "MI:0335" ) # certain
                        frange["end"]={ "position": rend }                                       

                elif ln.startswith("xref"):                    
                    db = col[1]
                    acc = col[2]
                    if len(col) > 3:
                        xtype = col[3]
                    else:
                        xtype = "identity"

                    if len(col) > 4:
                        dbac = col[4]
                    else:
                        dbac = self.cvdict[db]

                    if len(col) > 5:
                        xtypeac = col[5]
                    else:
                        xtypeac = self.cvdict[xtype]                    
                    
                    ver = None
                    if '.' in acc:
                        ver = acc[acc.rfind('.')+1:]
                        acc = acc[0:acc.rfind('.')]
                        if len(ver) == 0:
                            ver = None

                    if col[0].endswith(".p"):
                        xtgt = record["participant"]
                    
                    xref =  self.buildXref( acc, db=db, dbAc=dbac, version = ver,
                                            refType=xtype, refTypeAc=xtypeac )

                    if "xref" not in xtgt:                    
                        xtgt["xref"] = {}

                    if "primaryRef" not in xtgt["xref"]:
                        xtgt["xref"]["primaryRef"] = xref
                    else:
                        if "secondaryRef" not in xtgt["xref"]:
                            xtgt["xref"]["secondaryRef"] = []
                        xtgt["xref"]["secondaryRef"].append(xref)
                        
                elif ln.startswith("figure"):                
                    flabel = col[1]
                    if "atrribute" not in interaction:
                        interaction["attribute"]= []
                    att = {"name":"figure legend", "nameAc":"MI:0599","value":flabel}
                    interaction["attribute"].append(att)

                elif ln.startswith("seq"):                    
                    seq = col[1].upper()
                    if len(col) > 2:
                        nseq = col[2].upper()           
            
                    if "startStatus" in seqtgt:  # seqtgt is range                    
                        seqtgt["sequence"] = {}
                        seqtgt["sequence"]["original"]=seq
                        seqtgt["sequence"]["new"]=nseq
                        
                    else:
                        seqtgt["sequence"] = seq
                                            
                elif ln.startswith("#"):
                    self.feature = False
                    interaction = None
                    participant = None
                    feature = None
                    frange = None
                    seqtgt = None

        return pymex.mif.Record( {"entrySet":{ "entry":[record] } } )



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 02:05:33 2020

@author: andrei
"""

import json
import re

class Alias():

    def __init__(self, alias ):
        self._alias = alias
    
    def __repr__(self):
        return json.dumps(self._alias)

    def __str__(self):
        return str(self._alias["value"])

    @property
    def type(self):
        if "type" in self._alias:
            return self._alias["type"]
        return None
    
    @property
    def typeAc(self):
        if "typeAc" in self._alias:
            return self._alias["typeAc"] 
        return None


class Names():
    
    def __init__(self):
        pass
    
    def __repr__(self):
        return json.dumps(self._names)
    
    @property
    def label(self):
        return self._names["shortLabel"]
            
    @property
    def name(self):
        return self._names["fullName"]
    
    @property
    def alias(self):
        ret = []
        if "alias" in self._names:
            for a in self._names["alias"]:
                ret.append( Alias( a ) )
        return ret    
    
    
class Xref():
    """Xref representation."""
    def __init__( self ):        
        pass
    
    def __len__(self):
        if self._xref is None:
            return 0
        if "secondaryRef" in  self._xref:
            return 1 + len(self._xref["secondaryRef"])
        return 1
        
    @property
    def primaryRef(self):
        if self._xref is None:
            return None
        return Reference( self._xref["primaryRef"] )
    
    @property
    def secondaryRef(self):

        if self._xref is None:
            return None
        ret =[]
        
        if "secondaryRef" not in self._xref:
            return ret                
        
        for sr in self._xref["secondaryRef"]:
            ret.append( Reference(sr) )
        return ret
    
    @property
    def xrefs(self):
        if self._xref is None:
            return None
        ret = []
        ret.append( Reference(self._xref["primaryRef"]) )
        
        if  "secondaryRef" in self._xref:
            for sr in self._xref["secondaryRef"]:
                ret.append( Reference( sr ) )
        return ret
    
    @property
    def xrefCount(self):
        if self._xref is None:
            return 0
        if "secondaryRef" in self._xref:
            return 1 + len(self._xref["secondaryRef"])
        else:       
            return 1
             
class Bibref(Xref):
    def __init__( self, xref, attr ):
        self._xref = xref
        self._attribute = attr
                
    @property
    def attribs(self):    
        if self._attribute is None:
            return None
        return Attribs( self._attribute )
    
    
class Reference():
    def __init__(self, xref):
        self._xref = xref
           
    def __repr__(self):
        return json.dumps( self._xref )
        
    @property
    def db(self):
        if "db" in self._xref:
            return self._xref["db"]
        return None
    
    @property
    def dbAc(self):
        if "dbAc" in self._xref:
            return self._xref["dbAc"]
        return None

    @property
    def ac(self):
        if "id" in self._xref:
            return self._xref["id"]

        if "ac" in self._xref:
            return self._xref["ac"]

        return None
    
    @property
    def version(self):
        if "version" in self._xref:
            return self._xref["version"]
        return None
    
    @property
    def refType(self):
        if "refType" in self._xref:
            return self._xref["refType"]
        return None
    
    @property
    def refTypeAc(self):
        if "refTypeAc" in self._xref:
            return self._xref["refTypeAc"]
        return None



class Attribute():
    """MIF Attribute representation."""
    def __init__(self,attr):    
        self._attr = attr

    def __repr__(self):
        return json.dumps(self._attr)

    @property
    def name(self):
        if "name" in self._attr:
            return str(self._attr["name"])
        return None
    
    @property        
    def nameAc(self):
        if "nameAc" in self._attr:
            return str(self._attr["nameAc"])
        return None
     
    @property        
    def value(self):
        if "value" in self._attr:
            return str(self._attr["value"])
        return None
      

class Attribs():   
    def __init__(self, attribs):
        self._attribs = attribs
        
    def __getitem__(self, pos):    
        return Attribute( self._attribs[pos] )
   
    def __len__(self):
        return len(self._attribs)
    
    
class Parameter():
    """MIF Parameter representation."""
    def __init__(self, param):    
        self._param = param

    @property
    def term(self):
        return str(self._param["term"])

    @property
    def termAc(self):
        if "termAc" in self._param:
            return str(self._param["termAc"])
        return None
        
    @property
    def unit(self):
        if "unit" in self._param:
            return str(self._param["unit"])
        return None
    
    @property
    def unitAc(self):
        if "termAc" in self._param:
            return str(self._param["termAc"])
        return None

    @property
    def value(self):
        base = 10.0
        exp = 0.0
        factor = float(self._param["factor"])

        if "base" in self._param:
            base = float(self._param["base"])
        if "exponent" in self._param:
            exp = float(self._param["exponent"])   

        return factor * pow( base, exp)
            
    @property
    def uncertainty(self):
        if "uncertainty" in self._param:
            return float(self._param["uncertainty"])
        else:
            return float(0.0)
        
    @property
    def error(self):
        if "uncertainty" in self._param:
            return float(self._param["uncertainty"])
        else:
            return float(0.0)


class Params():    
    def __init__(self, params):
        self._params =  params
        
    def __getitem__(self, pos):    
        return Parameter( self._params[pos] )
   
    def __len__(self):
        return len(self._params)
    
    
class Source( Names, Xref ):
    """MIF Source representation."""   
    def __init__( self, source ):    
        self._source = source
        self._names = source["names"]
        
        self._xref = source["xref"] if "xref" in  source else None
        self._bibref = source["bibref"] if "bibref" in  source else None
        self._attribute = source["attribute"] if "attribute" in  source else None
                  

    @property
    def releasedate(self):
        if "releaseDate" in self._source:
            return str(self._source["releaseDate"])
        return None
            
    @property
    def bibref(self):    
        if self._bibref is not None:                           
            xref = self._bibref["xref"] if "xref" in self._bibref else None
            attr = self._bibref["attribute"] if "attribute" in self._bibref else None                 
            return Bibref( xref, attr )
        return None
    
    @property
    def attribs(self):    
        if self._attribute is not None:        
            return Attribs( self._attribute )
        return None
    
class Participant( Names,Xref ):
    """MIF Participant representation."""
    def __init__(self, participant, interaction ):    
        self._participant = participant
        self._interactor = self._participant["interactor"]
        self._interaction = interaction
        
        if "names" in  self._participant:            
            self._names = self._participant["names"]
        else:
            self._names = self._interactor["names"]
                
        self._xref = self._participant["xref"] if "xref" in self._participant else None
            
        if "hostOrganism" in self._participant:
            self._host = self._participant["hostOrganism"]
        else:
            self._host =  [ self._interactor["organism"] ]
            
        self._meth = None
        if "participantIdentificationMethod" in participant:
            self._meth = participant["participantIdentificationMethod"]
        else:
            if ("experiment" in interaction and 
                len(interaction["experiment"]) == 1):       
                ex = interaction["experiment"][0]
                if "participantIdentificationMethod" in ex:
                    self._meth =  ex["participantIdentificationMethod"]
            
        self._biorole = None
        if "biologicalRole" in participant:     
            if isinstance( participant["biologicalRole"], dict ):
                self._biorole = [participant["biologicalRole"]]
            elif isinstance( participant["biologicalRole"], list ):   
                self._biorole = participant["biologicalRole"]
                
        self._exprole = None
        if "experimentalRole" in participant:     
            if isinstance( participant["experimentalRole"], list ):
                self._exprole = participant["experimentalRole"]
            elif isinstance( participant["experimentalRole"], dict ):   
                self._exprole = [participant["experimentalRole"]]
             
        self._feature = self._participant["feature"] if "feature" in self._participant else None 
        self._confidence =self._participant["confidence"] if "confidence" in self._participant else None 
        self._parameter = self._participant["parameter"] if "parameter" in self._participant else None         
        self._attribute = self._participant["attribute"] if "attribute" in self._participant else None
            
    @property
    def interactor( self ):       
        return Interactor( self._interactor )
   
    @property
    def hosts( self ):
        ret = []
        for h in self._host: 
            ret.append( Host( h ) )        
        return ret
    
    @property
    def host( self ):
           if len( self._host ) == 1:
               return Host( self._host[0] )
           else:
               return None

    @property
    def meth( self ):
        if self._meth is not None:
            ret = []
            for m in self._meth:
                ret.append( CvTerm( m ) )
            return ret    
        return None

    @property
    def features( self ):
        ret = []
        for f in self._feature:
            ret.append( Feature( f ) )
            
        return ret
    
    @property
    def stoich(self):        
        #mif300 style: fixed value
        if "stoichiometry" in self._participant:
             sval = self._participant["stoichiometry"]
             try:                        
                 valb = float(sval)
                 vale = float(sval)
                 return (valb,vale)
             except:
                 return (str(sval),str(sval))
        #mif300 style: range
        if "stoichiometryRange" in self._participant:
             try:
                svalb = self._participant["stoichiometryRange"]["minValue"]
                svale = self._participant["stoichiometryRange"]["maxValue"]
                                                      
                return (float(svalb),float(svale))
             except:
                return (str(svalb),str(svale))             
            
        # mif254 style
        if "attribute" in self._participant:
            for a in self._participant["attribute"]:
                if "value" in a:
                    match = re.match('Stoichiometry: (.+)',a["value"])
                    if match:
                        sval = match.group(1)
                        try:                        
                            valb = float(sval)
                            vale = float(sval)
                            return (valb,vale)
                        except:
                            return (str(sval),str(sval))
                        
                    match = re.match('StoichiometryRange: (.+)',a["value"])
                    if match:
                        svalc = match.group(1).split(":")
                        try:                        
                            valb = float(svalc[0])
                            vale = float(svalc[1])
                            return (valb,vale)
                        except:
                            return (str(sval),str(sval))    
                                                
        return (0.0, 0.0)            
    
    @property
    def confidence( self ):
        if self._confidence is not None:
            return list(self._confidence)
    
    
class Experiment(Names, Xref):
    """MIF Experiment representation."""
    def __init__( self, experiment ):    
        self._experiment = experiment
                 
        self._names = self._experiment["names"] if "names" in self._experiment else None
        self._xref = self._experiment["xref"] if "xref" in self._experiment else None
        self._bibref = self._experiment["bibref"] if "bibref" in self._experiment else None
        self._attribute = self._experiment["attribute"] if "attribute" in self._experiment else None
          
        if "hostOrganism" in self._experiment:
            self._host = self._experiment["hostOrganism"]
        else:
            self._host = None
            
        self._meth = None
        if "interactionDetectionMethod" in self._experiment:
            self._meth = self._experiment["interactionDetectionMethod"]
       
        self._pmeth = None
        if "participantIdentificationMethod" in self._experiment:
            self._pmeth = self._experiment["participantIdentificationMethod"]

        self._fmeth = None
        if "featureIdentificationMethod" in self._experiment:
            self._fmeth = self._experiment["featureIdentificationMethod"]

    @property 
    def bibref(self):
        if self._bibref is not None:
            xref = self._bibref["xref"] if "xref" in self._bibref else None
            attr = self._bibref["attribute"] if "attribute" in self._bibref else None
                                         
            return Bibref( xref, attr )            
        return None
    
    @property
    def hosts( self ):
        if self._host is None:
            return None
        ret = []
        for h in self._host: 
            ret.append( Host( h ) )        
        return ret
    
    @property
    def host( self ):
           if len( self._host ) == 1:
               return Host( self._host[0] )
           else:
               return None
    
    @property   
    def method(self):
        if self._meth is not None:
            return CvTerm( self._meth )    
        return None
    
    @property       
    def partmethod(self):
        if self._pmeth is not None:
            return CvTerm( self._pmeth )
        return None
    
    @property        
    def featmethod(self):
        if self._fmeth is not None:
            return CvTerm( self._fmeth )        
        return None
    
    @property        
    def confidence(self):
        if self._conf is not  None:
            ret = []
            for c in self._conf:
                ret.append( Confidence( c ) )
            return ret 
        return None
        
    @property
    def attribs(self):    
        if self._attribute is not None:
            return Attribs( self._attribute )
        return None

class Interactor(Names, Xref):
    """MIF Interactor representation."""
    def __init__( self, interactor ):
        self._interactor = interactor
        self._names = interactor["names"]
        self._xref = interactor["xref"]
        self._type = interactor["interactorType"]
        self._host = interactor["organism"]
        
        self._sequence = None
        if "sequence" in interactor:
            self._sequence = interactor["sequence"]

    @property
    def type( self ):        
        return CvTerm(self._type)      

    @property
    def host( self ):        
        return Host( self._host )
    
    @property
    def sequence( self ):        
        return self._sequence
      

class Feature(Names, Xref):
    """MIF Feature representation."""
    def __init__( self, feature ):    
        self._feature = feature
        self._names= feature["names"]
        self._xref = feature["xref"]
        self._range =  feature["featureRange"]       
        
        self._meth = None
        if "featureDetectionMethod" in feature:
            self._meth = feature["featureDetectionMethod"]

        if "attribute" in feature:
            self._attribute = feature["attribute"]

    @property
    def type(self):
        return CvTerm(self._feature["featureType"])

    @property
    def methods( self ):
        if self._meth is not None:
            ret = []
            for m in self._meth:
                ret.append( CvTerm( m ) )
            return ret    
        return None
      
    
    @property
    def ranges(self):         
        ret = []
        for r in self._range:
            ret.append( Range(r) )
        return ret
    
    @property
    def attribs(self):    
        if self._attribute is not None:
            return Attribs( self._attribute )
        return None

class Range():
    def __init__(self, rng):
        self._rng = rng    

    @property
    def begStat(self):
        return CvTerm(self._rng["beginStatus"])

    @property
    def begPosition(self):
        if "begin" in self._rng:
            if "value" in self._rng["begin"]:
                pos = self._rng["begin"]["value"]
                return (int(pos),int(pos))
            if ("begin" in self._rng["begin"]  and 
                        "end" in self._rng["begin"]):
                posb = self._rng["begin"]["begin"]
                pose = self._rng["begin"]["end"]
                return ( int(posb) , int(pose) )    
        return None
       
    @property
    def endStat(self):
        return CvTerm(self._rng["endStatus"])

    @property
    def endPosition(self):
        if "begin" in self._rng:
            if "value" in self._rng["begin"]:
                pos = self._rng["begin"]["value"]
                return ( int(pos), int(pos) )
            if ("begin" in self._rng["begin"]  and 
                "end" in self._rng["begin"]):
                posb = self._rng["begin"]["begin"]
                pose = self._rng["begin"]["end"] 
                return ( int(posb), int(pose) )    
        return None
       
 
class Availability():
    """MIF Availability representation."""
    def __init__( self, avail ):
        self._avail = avail

    def __repr__(self):
        return str(self._avail)

        
class CvTerm( Names, Xref ):
    """Mif CvTerm representation."""
    def __init__(self, cvterm):        
        self._cvterm = cvterm
        self._names = cvterm["names"]
        self._xref = cvterm["xref"]

    def __repr__(self):
        return json.dumps(self._cvterm)


class Host( Names ):
    """MIF Host representation. Optional cell line/compartment/tissue information."""
    def __init__(self, host):
        self._host = host
        self._names = host["names"]
        
    def __repr__(self):
        return json.dumps(self._host)

    @property
    def taxid(self):
        return str(self._host["ncbiTaxId"])
    
    @property
    def cellType(self):
        if "cellType" in self._host:
            return CvTerm(self._host["cellType"])
        return None

    @property
    def compartment(self):
        if "compartment" in self._host:
            return CvTerm(self._host["compartment"])
        return None

    @property
    def tissue(self):
        if "tissue" in self._host:
            return CvTerm(self._host["tissue"])
        return None

class Confidence:
    """MIF Confidence representation. """
    def __init__(self, conf):
        self._conf = conf

    @property
    def unit(self):
        return CvTerm(self._conf["unit"])

    @property
    def value(self):
        return str(self._conf["value"])
    



#utf: # -*- coding-8 -*-
"""
Created on Mon Jun 29 13:56:57 2020
@author: andrei
"""

from lxml import etree as ET
import json
import os

class XmlRecord():
    """XML-based record representation."""
    
    def modifyTag(self,item,ver): 
        """ Modifies tag of an item if necessary."""
        tag = item.tag[self.config[ver]["NSL"]:]
        return tag
    
    def toNsString(self,ns, prefix=None): 
        """Converts namespace to string prefix for element tags."""
        mif_ns = ns[prefix]
        mifstr = "{%s}" % mif_ns
        return mifstr

    def __init__(self, root=None, config=None):

        if root is not None:
            self.root = root
        else:
            self.root = {}

        self.config = {}
        if config is not None:
            
            for ver in config.keys():
                self.config[ver] = {}
                
                self.config[ver]["IN"] = json.load( open( config[ver]["IN"] ) )
                self.config[ver]["NSL"] = len(self.config[ver]["IN"]["@NS"])+2
        
                self.config[ver]["OUT"] = json.load( open( config[ver]["OUT"] ) )

                # re-key  default ("*") namespace
                
                defns = None
                for nk in self.config[ver]["OUT"]["@NS"]:
                    if nk == "*":
                        defns = self.config[ver]["OUT"]["@NS"]["*"]
                if defns is not None:
                    self.config[ver]["OUT"]["@NS"].pop("*", None)
                    self.config[ver]["OUT"]["@NS"][None] = defns        
                
    def parseXml(self, filename, ver, debug=False):
        
        template = self.config[ver]["IN"]
        
        recordTree = ET.parse( filename )
        rootElem = recordTree.getroot()
        
        self.genericParse( template, ver, self.root, [] , rootElem, debug )
        return self

    def genericParse(self, template, ver, rec, rpath, elem, wrapped=False, debug=False):
      
        tag = self.modifyTag( elem, ver )

        #find corresponding template
        if tag in template:
            ttempl = template[tag]
        else:
            ttempl = template["*"] 

        if debug:
            print("\nTAG", tag, wrapped, len(rpath) )
            print(" TTEM", ttempl)
        
        if elem.getparent() is not None:
            parentElem = elem.getparent()
            if wrapped:
                if  parentElem.getparent() is not None:
                    parentElem = parentElem.getparent()
                else:
                    parentElem = None            
        else:
            parentElem = None
            
        if parentElem is not None:
            parentTag = self.modifyTag( parentElem, ver )
            if debug:
                print(" PAR", parentTag )
            
        else:
            parentTag = None
            if debug:
                print("PAR:  ROOT ELEM !!!")
        #print(parentTag)
        if parentTag is not None and parentTag in ttempl:
            ctempl = ttempl[parentTag]
        else:
            ctempl = ttempl["*"]
        if debug:
            print("  CTEMPL", ctempl )

        # find wrap flag
        if "wrapper" in ctempl and ctempl["wrapper"] is not None:
            cwrap = ctempl["wrapper"]
        else:
            #default
            cwrap = template["*"]["*"]["wrapper"]            
        if debug:
            print( "  CWRAP ", cwrap )

        if cwrap:
            for cchld in elem:
                if debug:
                    print("  CHLD",cchld.tag);

                self.genericParse( template, ver, rec, rpath, cchld, wrapped =True)
                if debug:
                    print( json.dumps(self.root, indent=2) )                
            return 
        
        # find current key:        
        if "name" in ctempl and ctempl["name"] is not None:
            ckey = ctempl["name"]
        else:
            ckey = tag
        
        # find complex flag
        if "complex" in ctempl and ctempl["complex"] is not None:
            ccomplex = ctempl["complex"]        
        else:            
            #default
            ccomplex = template["*"]["*"]["complex"]

        # test if reference
        rtgt  = None
        if "reference" in ctempl and ctempl["reference"] is not None:
             rtgt = ctempl["reference"]
        else:
            #default
            rtgt = template["*"]["*"]["reference"]

        # find current store type (direct/list/index)
        if "store" in ctempl and ctempl["store"] is not None:
            cstore = ctempl["store"]
        else:
            #default
            cstore = template["*"]["*"]["store"]
             
        if debug:
            print( "  CKEY  ", ckey )
            print( "  CCMPLX", ccomplex )
            print( "  CSTORE", cstore )
            print( "  CREFTG", rtgt )
            
        if rtgt is not None:
            # add referenced data
            # elem.text: a reference
            # rtgt     : path to referenced dictionary along current path
            #            within record data structure  

            stgt = rtgt.split('/')
            for i in range(1,len(stgt)):
                if stgt[i] in rpath[i-1]:
                    pass
                else:
                    break

            lastmatch = rpath[i-2][stgt[i-1]]
            if cstore == "list":
                if ckey not in rec:
                    rec[ckey] = []
                rec[ckey].append( lastmatch[stgt[i]][elem.text] )
            elif cstore == "direct":
                rec[ckey] = lastmatch[stgt[i]][elem.text]
                
        else:
            # build/add current value
            cvalue = None
            if ccomplex:
                cvalue = {}
                if elem.text:
                    val = str(elem.text).strip()
                    if len(val) > 0:
                        cvalue["value"] = val
            else:
                cvalue = str( elem.text )

            #if ckey in rec:
            #    print(ckey, rec[ckey])
                
            if cstore  == "direct":
                rec[ckey] = cvalue
            elif cstore == "list": 
                if ckey not in rec:
                    rec[ckey] = []
                rec[ckey].append(cvalue)
            elif cstore == "index":
                if ckey not in rec:
                    rec[ckey] = {}

                if "ikey" in ctempl and ctempl["ikey"] is not None:
                    ikey = ctempl["ikey"]
                else:
                    #default
                    ikey = template["*"]["*"]["ikey"]

                if ikey.startswith("@"):
                    iattr= ikey[1:]
                    ikeyval = elem.get(iattr)
                
                    if ikeyval is not None:                
                        rec[ckey][ikeyval] = cvalue        
            
            #create xrefInd inside xref
            if "index" in ctempl and ctempl["index"] is not None:
                ckeyRefInd = ctempl["index"]["name"]
                rec[ckey][ckeyRefInd] = {}
                
            # parse elem contents
                
            # add dom attributes
            for cattr in elem.attrib:
                cvalue[cattr] = str( elem.attrib[cattr] )

            cpath = []
            for p in rpath:
                cpath.append(p)
            
            cpath.append( {tag: cvalue })

            if 'index' in ctempl:
                iname = ctempl["index"]["name"]
                ientry = ctempl["index"]["entry"]                
                
            for cchld in elem:
                cchldTag = self.modifyTag(cchld, ver)
                if debug:
                    print("  CHLD", cchld.tag);
                    
                self.genericParse( template, ver, cvalue, cpath, cchld)
                
                if 'index' in ctempl:
                    if cchldTag in ientry and ientry[cchldTag] is not None:
                        keyList = ientry[cchldTag]["key"]

                        kval = []
                        for k in keyList:
                            kvl = cchld.xpath(k) 
                            if kvl:
                                kval.append(kvl[0])                        
                        dbkey = ':'.join(kval)
                        rec[ckey][ckeyRefInd][dbkey] = cvalue[cchldTag][0] if type(cvalue[cchldTag]) is list else cvalue[cchldTag]
                    

            if debug:
                print( json.dumps(self.root, indent=2) )
        return
        
    def parseJson(self, file ):
        self.root = json.load( file )
        return self

    def toJson(self):
        return json.dumps(self.root, indent=2)

    def toXml( self, ver='mif254', rdata = "entrySet", rtype="ExpandedEntrySet" ):
        """Builds Xml elementTree from a Record object."""

        template = self.config[ver]["OUT"]
        
        #nsmap = self.MIFNS[ver]
        nsmap = template["@NS"]

        return self.mifGenerator( nsmap, ver, template, None,
                                  self.root[rdata], template[rtype] )
        return None
    
    def mifGenerator(self, nsmap, ver, template, dom, cdata, ctype):
        """Returns DOM representation of the MIF record defined by the template.
        """
        
        self.UID = 1
        
        name = template["@ROOT"]["name"]
        if "attribute" in template["@ROOT"]:
            attrib = template["@ROOT"]["attribute"]
        else:
            attrib = None
            
        #dom = ET.Element(self.toNsString(self.MIFNS[ver]) + name,nsmap=nsmap)
        dom = ET.Element(self.toNsString(nsmap) + name,nsmap=nsmap)
        if attrib is not None:
            for a in attrib:

                if ":" in a:
                    (ans,aname)  = a.split(":")
                    qname = self.toNsString(nsmap,prefix=ans) + aname
                else:
                    qname = a
                    
                dom.set(qname,attrib[a])
            
        for cdef in ctype:            
            chldLst = self.mifElement( nsmap, ver, template, None, cdata, cdef)
            if chldLst is not None:
                for chld in chldLst:
                    dom.append( chld )
            
        return dom
    
    def mifElement(self, nsmap, ver, template, celem, cdata, cdef ):
        """Returns a list of DOM elements to be added to the parent and/or decorates  
           parent with attributes and text value . 
        """
        retLst = []
        if "wrap" in cdef:
            # add wrapper            
            wrapName = cdef["wrap"]
            wrapElem = ET.Element(self.toNsString(nsmap) + wrapName)        
        else:
            wrapElem = None
            
        if "value" not in cdef: # empty wrapper        
            # wrapper contents definition
            wrappedTypes = template[cdef["type"]]
            
            empty = True # flag: will skip if nothing inside 
            for wtDef in wrappedTypes: # go over wrapper content
                
                if wtDef["value"] in cdata:   # check if data present 
                    wElemData = cdata[wtDef["value"]]
                    for wed in wElemData: # wrapper contents *must* be a list
                        empty = False # non empty contents
                        
                        if "name" in wtDef:
                            chldName = wtDef["name"]
                        else:
                            chldName = wtDef["value"]

                        # create content element inside a wrapper
                        chldElem = ET.Element(self.toNsString(nsmap) + chldName)   
                        wrapElem.append(chldElem)

                        # fill in according to element definition
                        for wtp in template[wtDef["type"]]:
                            wclLst = self.mifElement( nsmap, ver, template,
                                                      chldElem, wed, wtp)                        
                            # add contents
                            if wclLst is not None:
                                for wcd in wclLst:  
                                    chldElem.append(wcd)
                                              
            if not empty:                                  
                return [wrapElem]
            else:
                return None
        
        if cdef["value"] =="$UID": #  generate next id
            self.UID += 1
            elemData = str(self.UID) 
        else:
            if cdef["value"] in cdata: # data from the record 
                elemData = cdata[ cdef["value"] ]
            else:               
                return [] # no data present: return empty list                   

        if "name" in cdef: # if present use name definition 
            elemName = cdef["name"]
        else:              # otherwise use the name of record field 
            elemName = cdef["value"]

            
        if isinstance( elemData, str):  # record field: text

            # attribute or text of the parent element

            if elemName.startswith('@'):  
                if elemName == "@TEXT" : # text of parent                
                    celem.text = str(elemData)
                else:                    # attribute of parent
                    celem.set(elemName[1:], str(elemData))                
            else: 
                if cdef["type"]=='$TEXT': # child element with text only                    
                    chldElem = ET.Element(self.toNsString(nsmap) + elemName)
                    chldElem.text = str( elemData )                

                    if wrapElem is not None: # add to wrapper (if present) 
                        wrapElem.append(chldElem)
                        return [wrapElem]
                    else: # otherwise add to return list            
                        retLst.append(chldElem)
            return retLst
                    
        if isinstance( elemData, dict): # record field: complex element 
            # convert single value to list
            elemData = [ elemData ]
             
        for celemData in elemData: # always a list here with one or more dict inside                    
            
            chldElem = ET.Element(self.toNsString(nsmap) + elemName)
            
            if cdef["type"]=='$TEXT': # text child element                 
                chldElem = ET.Element(self.toNsString(nsmap) + elemName)
                chldElem.text = str( celemData )
                retLst.append(chldElem)
            else: # complex content: build recursively                
                contType = template[cdef["type"]]
                           
                for contDef in contType: # go over definitions
                    cldLst = self.mifElement( nsmap, ver, template,
                                              chldElem,
                                              celemData,
                                              contDef)                
                    if cldLst is not None:
                        for cld in cldLst:  
                            chldElem.append( cld )
                    else: 
                        cldLst = None
                         
                if wrapElem is not None:       # if present, add to wrapper
                    if chldElem is not None:
                        wrapElem.append( chldElem )
                    else:
                       wrapElem = None 
                elif chldElem is not None:     #  otherwise add to return list    
                    retLst.append( chldElem )

        if wrapElem is not None and len(elemData) > 0:            
            return [wrapElem]
        
        return retLst    
    



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 02:05:33 2020

@author: andrei
"""

import os
import pymex
from pymex import xmlrecord, mif

class Record(xmlrecord.XmlRecord):
    """MIF record representation. Inherits XML parsing and serialization from xml.XmlRecord"""
    
    def __init__(self, root=None):

        myDir = os.path.dirname( os.path.realpath(__file__))
        self.mifConfig = { "mif254": {"IN": os.path.join( myDir, "defParse254.json"),
                                      "OUT": os.path.join( myDir, "defMif254.json" )},
                           "mif300": {"IN": os.path.join( myDir, "defParse300.json"),
                                      "OUT": os.path.join( myDir, "defMif300.json" )}}
        
        super().__init__(root, config=self.mifConfig )
                
    def parseMif(self, filename, ver="mif254", debug=False):
        return self.parseXml( filename, ver=ver )
   
    def toMif( self, ver='mif254' ):
        """Builds MIF elementTree from a Record object."""
        
        self._stoichiometryConvert( ver )        
        
        return self.toXml( ver, "entrySet", "ExpandedEntrySet" )
        
    def _stoichiometryConvert(self, ver):
        
        for e in self.root["entrySet"]["entry"]:
            for i in e["interaction"]:
                for p in i["participant"]:
                    if ver == 'mif254': 
                        # find mif300 stoichiometry
                        stset = False
                        if "stoichiometry" in p:
                            stval = str( p["stoichiometry"]["value"] )
                            stset =True
                        else:
                            stval = None
                            
                        if "stoichiometryRange" in p:
                            stmin = str( p["stoichiometryRange"]["minValue"] )
                            stmax = str( p["stoichiometryRange"]["maxValue"] )
                            stset =True
                        else:
                            stmin = None
                            stmax = None
                       
                        if stset:   
                            # replace/add mif254 stoichiometry
                            if "attribute" not in p:
                                p["attribute"]={}
                            ast = None    
                            for a in p["attribute"]:
                                if ("value" in a and 
                                    a["value"].startswith("Stoichiometry:") ):
                                    ast = a
                            if ast is None:        
                                ast =  {'name': 'comment', 'nameAc': 'MI:0612'}
                                p["attribute"].append(ast)
                                 
                            if stval is not None:        
                                ast["value"] = "Stoichiometry: " + stval
                                                 
                            elif stmin is not None and stmax is not None:
                                strng = stmin + " : " + stmax
                                ast["value"] = "StoichiometryRange: " + strng
                                        
                    elif ver == 'mif300':
                    
                        # find mif254 stoichiometry
                        stval = None
                        stmin = None
                        stmax = None
                        if "attribute" in p:
                            for a in p["attribute"]:                                
                                if ("value" in a and 
                                    a["value"].startswith("Stoichiometry:") 
                                    ):
                                        vcol= a["value"].split(" ")
                                        stval = vcol[1]
                                        p["attribute"].remove( a )
                                        break
                                                                
                                if ("value" in a and 
                                    a["value"].startswith("StoichiometryRange:")
                                    ):
                                        vcol= a["value"].split(" ")
                                        stmin = vcol[1]
                                        stmax = vcol[3]
                                        p["attribute"].remove( a )
                                        break
                                        
                        # replace/add mif300 stoichiometry 
                        if stval is not None:  
                            st = p.setdefault("stoichiometry",{})
                            st["value"] =str( stval )
                            
                        else:
                            pass
                            #p.pop("stoichiometry", None)
                            
                        if stmin is not None and stmax is not None:
                            st["minValue"] = str(stmin)                            
                            st["maxValue"] = str(stmax)
                        else:
                            pass
                            #p.pop("stoichiometryRange", None)
                            
    @property
    def entry(self):
        """Returns the first (default) entry of the record"""        
        return mif.Entry( self.root['entrySet']['entry'][0] )
    
    @property
    def entryCount(self):
        return len( self.root['entrySet']['entry'])
    
    def getEntry( self, n = 0 ):
        "Returns i-th entry of the record."""
        if n < len( self.root['entrySet']['entry'] ):
            return mif.Entry( self.root['entrySet']['entry'][n] )
        else:
            return None
    
    @property          
    def interactions(self):
        """Returns interactions of the first (default) entry of the record."""                        
        return mif.Entry( self.root['entrySet']['entry'][0] ).interactions
    
    @property
    def interactionCount(self):
        return len( self.root['entrySet']['entry'][0]["interaction"]) 
    
    def getInteraction( self, n ):
        return mif.Interaction( self.root['entrySet']['entry'][0], n )
    
       
class Entry():
    """MIF Entry representation."""
    def __init__( self, entry ):        
        self._entry = entry    
    
    @property
    def interactions( self ):
        ret = []        
        for i in range( 0, len( self._entry["interaction" ]) ):        
            ret.append( mif.Interaction( self._entry, n=i ) )        
        return ret
    
    @property    
    def interactionCount( self ):
        return len( self._entry[ "interaction" ] )
    
    def getInteraction( self, n ):
        return mif.Interaction( self._entry , n )
  
    @property
    def abstIinteractions( self ):
        ret = []        
        for i in range( 0, len( self._entry["abstInteraction" ]) ):        
            ret.append( mif.AbstInteraction( self._entry, n=i ) )        
        return ret
    
    @property    
    def abstInteractionCount( self ):
        return len( self._entry[ "abstInteraction" ] )
    
    def getAbstInteraction( self, n ):
        return mif.AbstInteraction( self._entry , n )
    
    
    @property
    def source(self):
        return mif.Source(self._entry["source"])

       








        
        
        
        




# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 02:05:33 2020

@author: andrei
"""

import os
import sys

import json
import re
from copy import deepcopy

import pymex
from pymex import mif

class Interaction(mif.Names, mif.Xref):
    """MIF Interaction representation."""
    
    physical = {"names": {
                 "shortLabel": "physical association",
                 "fullName": "physical association"
                 },
              "xref": {
                  "primaryRef": {
                     "db": "psi-mi",
                     "dbAc": "MI:0488",
                     "id": "MI:0915",
                     "refType": "identity",
                     "refTypeAc": "MI:0356"
                     } 
                 }
             } 
         
    direct = {"names": {
                 "shortLabel": "direct interaction",
                 "fullName": "direct interaction"
                 },
              "xref": {
                  "primaryRef": {
                     "db": "psi-mi",
                     "dbAc": "MI:0488",
                     "id": "MI:0407",
                     "refType": "identity",
                     "refTypeAc": "MI:0356"
                     } 
                 }
             }           
            
    
    def __init__( self, entry, n, interaction=None ): 
        self._entry = entry
        if n is not None:
            self._interaction = entry[ "interaction" ][ n ]
        else:
            self._interaction = interaction
            
        self._source = entry[ "source" ]
        self._experiment =  self._interaction["experiment"]
        self._participant = self._interaction["participant"]

        self._names = self._interaction["names"] if "names" in self._interaction else None 
        self._xref = self._interaction["xref"] if "xref" in self._interaction else None
        self._attribute = self._interaction["attribute"] if "attribute" in self._interaction else None

            
    def expand(self, mode="spoke"):
        """Returns a list of binary interactions generated by expansion
        of multi-molecular interaction according to spoke or marix model.""" 
        
        ret = []
      
        if len(self._participant) <=2:
                return ret
      
        bait = []
        prey = []
        part = []
        
        assType = False
        phyType = False
        dirType = False
        
        if "interactionType" in self._interaction:
            for it in self._interaction["interactionType"]:
                if it["names"]["shortLabel"]=="association":
                    assType = True
                if it["names"]["shortLabel"]=="physical association":
                    phyType = True
                if it["names"]["shortLabel"]=="direct interaction":
                    dirType = True   
                    
        for p in self._participant:
            skip = False
            if "experimentalRole" in p:
                for r in p["experimentalRole"]:
                    if r["names"]["shortLabel"] == "bait":
                        bait.append(p)
                    if r["names"]["shortLabel"] == "prey":
                        prey.append(p)
                    if r["names"]["shortLabel"] == "ancillary":
                        skip = True
            else:
                skip = True 
            if not skip:            
                part.append(p)            
                
        if assType and mode == "spoke":
            for p1 in bait:
                for p2 in prey:
                                      
                    binary = {}
                    for k in self._interaction.keys():
                        if k not in ["participant","interactionType"]:
                            binary[k] = self._interaction[k]
                  
                    itype = deepcopy( self.physical )
                    binary.setdefault("interactionType",[]).append( itype )
                    bprt =  binary.setdefault("participant",[])
                    bprt.append(p1)
                    bprt.append(p2)
                    
                    ret.append( Interaction( self._entry, None, binary ) )
                    
        if (phyType or dirType) and mode == "matrix":           
            for i in range(0,len(part)):
                for j in range(i+1,len(part)):
                    binary = {}
                        
                    for k in self._interaction.keys():
                        if k not in ["participant","interactionType"]:
                            binary[k] = self._interaction[k]

                    if  dirType:                
                        itype = deepcopy( self.direct )
                    else:
                        itype = deepcopy( self.physical )
                    binary.setdefault("interactionType",[]).append( itype )
                    bprt =  binary.setdefault("participant",[])
                    bprt.append(part[i])
                    bprt.append(part[j])
                    
                    ret.append( Interaction( self._entry, None, binary ) )
                    
        return ret
        
    @property
    def imexid(self):
        if "imexId" in self._interaction:
            return self._interaction["imexId"]
        return None
    
    @property
    def participants(self):
        ret = []        
        for i in self._participant: 
            ret.append( mif.Participant( i,  self._interaction ) )        
        return ret
        
    @property 
    def participantCount(self): 
        return len( self._participant )

    def getParticipant( self, n = 0 ):
        if n < len( self._participant ):            
            return mif.Participant( self._participant[ n ] ) 
        else:    
            return None
        
    @property
    def type(self):
        if "interactionType" in self._interaction:
            if isinstance(self._interaction["interactionType"], list):            
                return mif.CvTerm(self._interaction[ "interactionType" ][0])
            if isinstance(self._interaction["interactionType"], dict):            
                return mif.CvTerm(self._interaction[ "interactionType" ])
        return None
    
    @property
    def types(self):
        if "interactionType" in self._interaction:
            if isinstance(self._interaction["interactionType"], list):
                ret = []
                for t in self._interaction["interactionType"]:
                    ret.append( mif.CvTerm(t) )
                return ret
        return None

    @property
    def source(self):
        return mif.Source( self._source )

    @property
    def experiments(self):
        ret = []
        for i in self._experiment:
            ret.append( mif.Experiment( i ) )
        return ret
        
    @property
    def availability(self):
        if "availability" in self._interaction:
            return self._interaction["availability"]["value"]
        return None
    
    @property 
    def experimentCount(self): 
        return len( self._experiment )
      
    @property
    def experiment(self):
        if len( self._experiment ) == 1:
            return mif.Experiment( self._experiment[0] )
        else:
            return None
    
    def getExperiment( self, n = 0 ):
        if n < len( self._experiment ):            
            return mif.Experiment( self._experiment[ n ] ) 
        else:    
            return None
                
    @property
    def attribs(self):    
        if self._attribute is None:
            return None
        return mif.Attribs( self._attribute )

    @property
    def params(self):    
        if self._param is None:
            return None
        return mif.Params( self._param )
       
    @property
    def confidence(self):
        if "confidence" in  self._interaction:
            return bool(self._interaction["confidence"])
    
    @property
    def modelled(self):
        if "modelled" in  self._interaction:
            return bool(self._interaction["modelled"])
        return False

    @property
    def intramolecular(self):
        if "intramolecular" in  self._interaction:
            return bool(self._interaction["intramolecular"])
        return False
    
    @property
    def negative(self):
        if "negative" in  self._interaction:
            return bool(self._interaction["negative"])
        return False
    

class AbstInteraction(mif.Names, mif.Xref):
    """MIF AbstractInteraction representation."""
    def __init__( self, entry, n ):    
        self._entry = entry
        self._interaction = entry[ "abstractInteraction" ][ n ]
        self._source = entry[ "source" ]       
        self._participant = self._interaction["participant"]

        self._names = self._interaction["names"] if "names" in self._interaction else None
        self._xref = self._interaction["xref"] if "xref" in self._interaction else None
        self._attribute = self._interaction["attribute"] if "attribute" in self._interaction else None
       
    @property
    def participants(self):
        ret = []        
        for i in self._participant: 
            ret.append( mif.Participant( i,  self._interaction ) )        
        return ret
        
    @property 
    def participantCount(self): 
        return len( self._participant )

    def getParticipant( self, n = 0 ):
        if n < len( self._participant ):            
            return mif.Participant( self._participant[ n ] ) 
        else:    
            return None
        
    @property
    def type(self):
        if "interactionType" in self._interaction:            
            if isinstance(self._interaction["interactionType"], dict):            
                return mif.CvTerm(self._interaction[ "interactionType" ])
        return None

    @property
    def interactorType(self):
        if "interactorType" in self._interaction:            
            if isinstance(self._interaction["interactorType"], dict):            
                return mif.CvTerm(self._interaction[ "interactorType" ])
        return None
    
    @property
    def evidenceType(self):
        if "evidenceType" in self._interaction:            
            if isinstance(self._interaction["evidenceType"], dict):            
                return mif.CvTerm(self._interaction[ "evidenceType" ])
        return None

    @property
    def organism(self):
        if "organism" in self._interaction:                                  
            return mif.Host( self._interaction["organism"] )
        return None 

    @property
    def source(self):
        return mif.Source( self._source )

    @property
    def attribs(self):    
        if self._attribute is not None:
            return mif.Attribs( self._attribute )
        return None
    
    @property
    def params(self):    
        if self._param is None:
            return None
        return mif.Params( self._param )
       







        
        
        
        




class Registry():

    def __init__( self, reg ):
        self.reg = reg

    def __str__( self ):
            return self.__repr__()
        
    def __repr__( self ):

        if self.reg is None:
            return ''

        repr = [ "PSICQUC Registry" ]
        for s in sorted(self.reg):
            if self.reg[s]['active'].lower() == 'true':
                status = 'active'
            else:
                status = 'inactive'

            repr.append( " Name: %-20s Status: %s " % ( self.reg[s]['name'] ,
                                                        status ) )
            repr.append( "  Rest URL: %s " %(self.reg[s]['rurl']))
        
        return '\n'.join(repr)
        



import setuptools
import os
from distutils.util import convert_path

ldesc =[]

with open( "README.md", "r") as fh:
        long_description = fh.read()
                
pkg_dir = convert_path('pylib/pymex')
version_file = open(os.path.join(pkg_dir, 'VERSION'))
version = version_file.read().strip()

setuptools.setup( name="pymex",
                  version=version,
                  author="Pymex Support Group",
                  author_email="dip@mbi.ucla.edu",
                  description="PYMEX: PSI-MI/IMEx Data Access",
                  long_description=long_description,
                  long_description_content_type='text/markdown',
                  url="https://github.com/IMExConsortium/pymex",
                  download_url = 'https://github.com/IMExConsortium/pymex/archive/'+str(version)+'.tar.gz',
                  package_dir={'': 'pylib'},                
                  package_data={
                          'pymex': ['mif/*.json', 'mif/*.jmif','mif/*.xsd'],
                  },
                  include_package_data=True,
                  packages=setuptools.find_packages(where="pylib",include=['pymex', 'pymex.*']),
                  classifiers=[
                          "Programming Language :: Python :: 3",
                          "License :: OSI Approved :: MIT License",
                          "Operating System :: OS Independent",
                          "Development Status :: 4 - Beta",
                          "Topic :: Scientific/Engineering :: Bio-Informatics",
                          "Intended Audience :: Science/Research"
                 ],
                 python_requires='>=3.5',
                 install_requires=[ "lxml>=3.5.0", "zeep>=3.4"]
)
